import pytest
import textwrap


@pytest.fixture()
def mermaid_full_string_with_no_layout(mermaid_full_string_preseve_column_sort: str) -> str:
    return mermaid_full_string_preseve_column_sort


@pytest.fixture()
def mermaid_full_string_with_dagre_layout(mermaid_full_string_preseve_column_sort: str) -> str:
    front_matter = textwrap.dedent(
        """
        ---
            config:
                layout: dagre
        ---
        """
    )
    return f"{front_matter}{mermaid_full_string_preseve_column_sort}"


@pytest.fixture()
def mermaid_full_string_with_elk_layout(mermaid_full_string_preseve_column_sort: str) -> str:
    front_matter = textwrap.dedent(
        """
        ---
            config:
                layout: elk
        ---
        """
    )
    return f"{front_matter}{mermaid_full_string_preseve_column_sort}"



def mermaid_assert(output: str) -> None:
    assert "users {" in output
    assert "posts {" in output
    assert "comments {" in output

    assert "users ||--o{ posts : author" in output
    assert "posts ||--o{ comments : post" in output
    assert "users ||--o{ comments : author" in output

    assert "CHAR(32) author FK" in output
    assert 'CHAR(32) post FK "nullable"' in output
    assert 'BOOLEAN live "True if post is published,nullable"' in output
    assert "DATETIME created" in output

    trailing_newline_assert(output)


def dot_assert(output: str) -> None:
    assert '<tr><td colspan="3" bgcolor="lightblue"><b>users</b></td></tr>' in output
    assert '<tr><td colspan="3" bgcolor="lightblue"><b>posts</b></td></tr>' in output
    assert '<tr><td colspan="3" bgcolor="lightblue"><b>comments</b></td></tr>' in output

    assert "users -- posts [label=author, dir=both, arrowhead=crow, arrowtail=none];" in output
    assert "posts -- comments [label=post, dir=both, arrowhead=crow, arrowtail=none];" in output
    assert "users -- comments [label=author, dir=both, arrowhead=crow, arrowtail=none];" in output

    assert '<tr><td align="left">CHAR(32)</td><td align="left">author</td><td>Foreign Key</td></tr>' in output
    assert '<tr><td align="left">CHAR(32)</td><td align="left">post</td><td>Foreign Key</td></tr>' in output
    assert '<tr><td align="left">DATETIME</td><td align="left">created</td><td></td></tr>' in output

    trailing_newline_assert(output)


def trailing_newline_assert(output: str) -> None:
    """
    Check that the output ends with a single newline.
    This reduces end user linter rewrites,
    e.g. from pre-commit's end-of-file-fixer hook.
    """
    assert output.endswith("\n")
    assert not output.endswith("\n\n")



import importlib
import os
import sys
from pathlib import Path
import re
from typing import List, Set, Optional, Dict, Union

from sqlalchemy.schema import MetaData

from .transformers.dot import Dot
from .transformers.mermaid import Mermaid

transformers: Dict[str, type[Union[Mermaid, Dot]]] = {
    "mmd": Mermaid,
    "mermaid": Mermaid,
    "dot": Dot,
    "gv": Dot,
}


def get_graph_string(
    *,
    base_class_path: str,
    import_module: List[str],
    include_tables: Set[str],
    exclude_tables: Set[str],
    python_dir: List[Path],
    format: str,
    column_sort: str,
    omit_comments: bool = False,
    layout: Optional[str] = None,
) -> str:
    # Update the PYTHON_PATH to allow more module imports.
    sys.path.append(str(os.getcwd()))
    for dir in python_dir:
        sys.path.append(str(dir))

    # Import the base class so the metadata class can be extracted from it.
    # The metadata class is passed to the transformer.
    module_path, class_name = base_class_path.split(":", 2)
    base_module = importlib.import_module(module_path)
    base_class = getattr(base_module, class_name)
    metadata = base_class.metadata

    # The modules holding the model classes have to be imported to get put in the metaclass model registry.
    # These modules aren't actually used in any way, so they are discarded.
    # They are also imported in scope of this function to prevent namespace pollution.
    for module in import_module:
        if ":*" in module:
            # Sure, execs are gross, but this is the only way to dynamically import wildcards.
            exec(f"from {module[:-2]} import *")
        else:
            importlib.import_module(module)

    # Grab a transformer.
    if format not in transformers:
        raise ValueError(f"Unknown Format: {format}")
    transformer = transformers[format]

    # Keep only the tables which were included / not-excluded
    include_tables = resolve_included_tables(
        include_tables=include_tables, exclude_tables=exclude_tables, all_tables=set(metadata.tables.keys())
    )
    filtered_metadata = filter_metadata(metadata=metadata, include_tables=include_tables)

    # Save the graph structure to string.
    return str(transformer(filtered_metadata, column_sort, omit_comments=omit_comments, layout=layout))


def resolve_included_tables(
    include_tables: Set[str],
    exclude_tables: Set[str],
    all_tables: Set[str],
) -> Set[str]:
    """Resolves the final set of tables to include in the graph.

    Given sets of inclusions and exclusions and the set of all tables we define
    the following cases are:
    - Empty inclusion and empty exclusion -> include all tables.
    - Empty inclusion and some exclusions -> include all tables except the ones in the exclusion set.
    - Some inclusions and empty exclusion -> make sure tables in the inclusion set are present in
        all tables then include the tables in the inclusion set.
    - Some inclusions and some exclusions -> not resolvable, an error is raised.
    """
    match len(include_tables), len(exclude_tables):
        case 0, 0:
            return all_tables
        case 0, int():
            excluded = {table for table in all_tables if any(re.match(pattern, table) for pattern in exclude_tables)}
            return all_tables - excluded
        case int(), 0:
            included = {table for table in all_tables if any(re.match(pattern, table) for pattern in include_tables)}

            if not included:
                non_existent_tables = include_tables - all_tables
                raise ValueError(
                    f"Some tables to include ({non_existent_tables}) don't exist"
                    "withinthe found tables ({all_tables})."
                )
            return included
        case _:
            raise ValueError(
                f"Only one or none of include_tables ({include_tables}) or exclude_tables"
                f"({exclude_tables}) can contain values."
            )


def filter_metadata(
    metadata: MetaData,
    include_tables: Set[str],
) -> MetaData:
    """Create a subset of the metadata based on the tables to include."""
    filtered_metadata = MetaData()
    for tablename, table in metadata.tables.items():
        if tablename in include_tables:
            if hasattr(table, "to_metadata"):
                # to_metadata is the new way to do this, but it's only available in newer versions of SQLAlchemy.
                table = table.to_metadata(filtered_metadata)
            else:
                # tometadata is deprecated, but we still need to support it for older versions of SQLAlchemy.
                table = table.tometadata(filtered_metadata)

    return filtered_metadata



from paracelsus.transformers.mermaid import Mermaid

from ..utils import mermaid_assert


def test_mermaid(metaclass):
    mermaid = Mermaid(metaclass=metaclass, column_sort="key-based")
    graph_string = str(mermaid)
    mermaid_assert(graph_string)


def test_mermaid_column_sort_preserve_order(metaclass, mermaid_full_string_preseve_column_sort):
    mermaid = Mermaid(metaclass=metaclass, column_sort="preserve-order")
    assert str(mermaid) == mermaid_full_string_preseve_column_sort


def test_mermaid_keep_comments(metaclass):
    mermaid = Mermaid(metaclass=metaclass, column_sort="key-based", omit_comments=False)
    assert "True if post is published" in str(mermaid)


def test_mermaid_omit_comments(metaclass):
    mermaid = Mermaid(metaclass=metaclass, column_sort="key-based", omit_comments=True)
    assert "True if post is published" not in str(mermaid)


def test_mermaid_with_no_layout(metaclass, mermaid_full_string_with_no_layout):
    mermaid = Mermaid(metaclass=metaclass, column_sort="preserve-order", layout=None)
    assert str(mermaid) == mermaid_full_string_with_no_layout


def test_mermaid_with_dagre_layout(metaclass, mermaid_full_string_with_dagre_layout):
    mermaid = Mermaid(metaclass=metaclass, column_sort="preserve-order", layout="dagre")
    assert str(mermaid) == mermaid_full_string_with_dagre_layout


def test_mermaid_with_elk_layout(metaclass, mermaid_full_string_with_elk_layout):
    mermaid = Mermaid(metaclass=metaclass, column_sort="preserve-order", layout="elk")
    assert str(mermaid) == mermaid_full_string_with_elk_layout



from pathlib import Path
from typing import Literal
import pytest
from unittest.mock import Mock
from unittest import mock

from typer.testing import CliRunner

from paracelsus.cli import app

from .utils import mermaid_assert

runner = CliRunner()


def test_graph(package_path: Path):
    result = runner.invoke(
        app,
        ["graph", "example.base:Base", "--import-module", "example.models", "--python-dir", str(package_path)],
    )

    assert result.exit_code == 0
    mermaid_assert(result.stdout)


@pytest.mark.parametrize("column_sort_arg", ["key-based", "preserve-order"])
def test_graph_column_sort(package_path: Path, column_sort_arg: Literal["key-based"] | Literal["preserve-order"]):
    result = runner.invoke(
        app,
        [
            "graph",
            "example.base:Base",
            "--import-module",
            "example.models",
            "--python-dir",
            str(package_path),
            "--column-sort",
            column_sort_arg,
        ],
    )

    assert result.exit_code == 0
    mermaid_assert(result.stdout)


def test_graph_with_exclusion(package_path: Path):
    result = runner.invoke(
        app,
        [
            "graph",
            "example.base:Base",
            "--import-module",
            "example.models",
            "--python-dir",
            str(package_path),
            "--exclude-tables",
            "comments",
        ],
    )
    assert result.exit_code == 0
    assert "posts {" in result.stdout
    assert "comments {" not in result.stdout


def test_graph_with_inclusion(package_path: Path):
    result = runner.invoke(
        app,
        [
            "graph",
            "example.base:Base",
            "--import-module",
            "example.models",
            "--python-dir",
            str(package_path),
            "--include-tables",
            "comments",
        ],
    )
    assert result.exit_code == 0
    assert "posts {" not in result.stdout
    assert "comments {" in result.stdout


def test_inject_check(package_path: Path):
    result = runner.invoke(
        app,
        [
            "inject",
            str(package_path / "README.md"),
            "example.base:Base",
            "--import-module",
            "example.models",
            "--python-dir",
            str(package_path),
            "--check",
        ],
    )
    assert result.exit_code == 1


def test_inject(package_path: Path):
    result = runner.invoke(
        app,
        [
            "inject",
            str(package_path / "README.md"),
            "example.base:Base",
            "--import-module",
            "example.models",
            "--python-dir",
            str(package_path),
        ],
    )
    assert result.exit_code == 0

    with open(package_path / "README.md") as fp:
        readme = fp.read()
        mermaid_assert(readme)


@pytest.mark.parametrize("column_sort_arg", ["key-based", "preserve-order"])
def test_inject_column_sort(package_path: Path, column_sort_arg: Literal["key-based"] | Literal["preserve-order"]):
    result = runner.invoke(
        app,
        [
            "inject",
            str(package_path / "README.md"),
            "example.base:Base",
            "--import-module",
            "example.models",
            "--python-dir",
            str(package_path),
            "--column-sort",
            column_sort_arg,
        ],
    )
    assert result.exit_code == 0

    with open(package_path / "README.md") as fp:
        readme = fp.read()
        mermaid_assert(readme)


@mock.patch(
    "paracelsus.cli.get_pyproject_settings", return_value={"base": "example.base:Base", "imports": ["example.models"]}
)
def test_inject_pyproject_configuration(mock_get_pyproject_settings: Mock, package_path: Path):
    """Test that the pyproject.toml configuration is used when base class path is not passed as argument."""
    result = runner.invoke(
        app,
        ["inject", str(package_path / "README.md")],
    )
    assert result.exit_code == 0

    with open(package_path / "README.md") as fp:
        readme = fp.read()
        mermaid_assert(readme)

    mock_get_pyproject_settings.assert_called_once()


def test_version():
    result = runner.invoke(app, ["version"])
    assert result.exit_code == 0


def test_graph_with_inclusion_regex(package_path: Path):
    result = runner.invoke(
        app,
        [
            "graph",
            "example.base:Base",
            "--import-module",
            "example.models",
            "--python-dir",
            str(package_path),
            "--include-tables",
            "^com.*",
        ],
    )
    assert result.exit_code == 0
    assert "comments {" in result.stdout
    assert "users {" not in result.stdout
    assert "post{" not in result.stdout


def test_graph_with_exclusion_regex(package_path: Path):
    result = runner.invoke(
        app,
        [
            "graph",
            "example.base:Base",
            "--import-module",
            "example.models",
            "--python-dir",
            str(package_path),
            "--exclude-tables",
            "^pos*.",
        ],
    )
    assert result.exit_code == 0
    assert "comments {" in result.stdout
    assert "users {" in result.stdout
    assert "post {" not in result.stdout


@pytest.mark.parametrize("layout_arg", ["dagre", "elk"])
def test_graph_layout(package_path: Path, layout_arg: Literal["dagre", "elk"]):
    result = runner.invoke(
        app,
        [
            "graph",
            "example.base:Base",
            "--import-module",
            "example.models",
            "--python-dir",
            str(package_path),
            "--layout",
            layout_arg,
        ],
    )

    assert result.exit_code == 0
    mermaid_assert(result.stdout)


@pytest.mark.parametrize("layout_arg", ["dagre", "elk"])
def test_inject_layout(package_path: Path, layout_arg: Literal["dagre", "elk"]):
    result = runner.invoke(
        app,
        [
            "inject",
            str(package_path / "README.md"),
            "example.base:Base",
            "--import-module",
            "example.models",
            "--python-dir",
            str(package_path),
            "--layout",
            layout_arg,
        ],
    )
    assert result.exit_code == 0

    with open(package_path / "README.md") as fp:
        readme = fp.read()
        mermaid_assert(readme)



from paracelsus.transformers.dot import Dot

from ..utils import dot_assert


def test_dot(metaclass):
    dot = Dot(metaclass=metaclass, column_sort="key-based")
    graph_string = str(dot)
    dot_assert(graph_string)


def test_dot_column_sort_preserve_order(metaclass, dot_full_string_preseve_column_sort):
    dot = Dot(metaclass=metaclass, column_sort="preserve-order")
    assert str(dot) == dot_full_string_preseve_column_sort
